
/* ================================================================

      File:  DVCLIP.C

   Purpose:  Contains routines related to cutting/pasting bitmaps to/from
             the clipboard.

 Functions:  CopyHandle
             CropBitmap
             RenderFormat
             HandleCopyClipboard
             HandlePasteClipboard

  Comments:  

   History:   Date      Reason
             6/ 1/91     Created

=================================================================== */
#include	"dv.h"		// Single inclusive include
//#include <windows.h>
//#include "DvChild.h"
//#include "DvClip.h"
//#include "DvDib.h"
//#include "DvErrors.h"
//#include "DvFrame.h"
//#include "DvPal.h"
//#include "DvMem.h"
// NEW
extern   PMWL	AddToFileList4( PRDIB prd );
extern   BOOL  WriteBMPFile2( LPTSTR lpf, HGLOBAL hDIB, BOOL bShow );

//RECT  grcClip     = {0,0,0,0};    // Clipboard rectangle at time of copy.
//DWORD		gnPasteNum = 1;      // For window title
TCHAR gszRPFrm[] = DEF_FM_STG;   // "TEMPB%03d.BMP" - when NO FILE, use this

//---------------------------------------------------------------------
// Function:   CopyHandle
// Purpose:    Makes a copy of the given global memory block.  Returns
//             a handle to the new memory block (NULL on error).
//             Routine stolen verbatim out of ShowDIB.
// Parms:      h == Handle to global memory to duplicate.
// Returns:    Handle to new global memory block.
// History:   Date      Reason
//             ???      Created
//---------------------------------------------------------------------
HANDLE CopyHandle( HANDLE h )
{
	PINT8	lpCopy;
	PINT8	lp;
	HANDLE     hCopy;
	DWORD      dwLen;

	hCopy = 0;
	if( ( h ) &&
		( dwLen = GlobalSize( h ) ) )
	{
		//if( hCopy = DVGlobalAlloc( GHND, dwLen ) )
		if( hCopy = DVGAlloc( "CopyHand", GHND, dwLen ) )
		{
			if( lpCopy = (PINT8)DVGlobalLock( hCopy ) )
         {
			   // NOTE: Could be a handle from CLIPBOARD
			   if( lp = (PINT8)GlobalLock( h ) )
            {
#ifdef	WIN32
			      memcpy( lpCopy, lp, dwLen );
#else	// !WIN32
			      while( dwLen-- )
				      *lpCopy++ = *lp++;	// Slow BYTE by BYTE copy
#endif	// WIn32 y/n
      			GlobalUnlock( h );	// Note: A Clipboard Handle, for ex.
            }

            DVGlobalUnlock( hCopy );
         }
		}
	}
	return( hCopy );
}

//---------------------------------------------------------------------
//
// Function:   CropBitmap
//
// Purpose:    Crops a bitmap to a new size specified by the lpRect
//             parameter.  The lpptSize parameter is used to determine
//             how much to stretch/compress the bitmap.  Returns a
//             handle to a new bitmap.  If lpRect is empty, copies the
//             bitmap to a new one.
//
//             Stolen almost verbatim out of ShowDIB.
//
// Parms:      hbm      == Handle to device dependent bitmap to crop.
//             hPal     == Palette to use in cropping (NULL for default pal.)
//             lpRect   == New bitmap's size (size we're cropping to).
//             lpptSize == A scaling factor scale by the proportion:
//                              Bitmap Width / lpptSize->x horizontally,
//                              Bitmap Height / lpptSize->y horizontally.
//                           Note that if lpptSize is set to the bitmap's
//                           dimensions, no scaling occurs.
//             
//
// History:   Date      Reason
//            6/15/91   Stolen from ShowDIB
//             
//---------------------------------------------------------------------

HBITMAP CropBitmap (HBITMAP hbm, 
                   HPALETTE hPal, 
                     LPRECT lpRect, 
                    LPPOINT lpptSize)
{
	HDC      hMemDCsrc;
	HDC      hMemDCdst;
	BITMAP   bm;
	int      dxDst,dyDst, dxSrc, dySrc;
	double   cxScale, cyScale;
	HPALETTE hOldPal1 = NULL;
	HPALETTE hOldPal2 = NULL;
	HBITMAP  hNewBm = NULL;

	if( !hbm )		// Oops, no BITMAP!!!
		return( hNewBm );


	GetObject( hbm, sizeof(BITMAP), (LPSTR)&bm );


	hMemDCsrc = CreateCompatibleDC( NULL );
	hMemDCdst = CreateCompatibleDC( NULL );


	if( hPal )
	{
		hOldPal1 = SelectPalette( hMemDCsrc, hPal, FALSE );
		hOldPal2 = SelectPalette( hMemDCdst, hPal, FALSE );
		RealizePalette( hMemDCdst );
	}


	dxDst     = lpRect->right  - lpRect->left;
	dyDst     = lpRect->bottom - lpRect->top;
	cxScale   = (double) bm.bmWidth  / lpptSize->x;
	cyScale   = (double) bm.bmHeight / lpptSize->y;
	dxSrc     = (int) ((lpRect->right - lpRect->left) * cxScale);
	dySrc     = (int) ((lpRect->bottom - lpRect->top) * cyScale);


	if( (dxDst == 0) || (dyDst == 0) )
	{
		dxDst = bm.bmWidth;
		dyDst = bm.bmHeight;
	}


	if( dxSrc == 0 )
		dxSrc = 1;

	if( dySrc == 0 )
		dySrc = 1;


	hNewBm = CreateBitmap( dxDst, dyDst, bm.bmPlanes, bm.bmBitsPixel, NULL );

	if( hNewBm )
	{
		HBITMAP hOldBitmap1, hOldBitmap2;

		hOldBitmap1 = SelectObject( hMemDCsrc, hbm );
		hOldBitmap2 = SelectObject( hMemDCdst, hNewBm );

		StretchBlt( hMemDCdst,
			0,
			0,
			dxDst,
			dyDst,
			hMemDCsrc,
			(int) (lpRect->left * cxScale),
			(int) (lpRect->top  * cyScale),
			dxSrc,
			dySrc,
			SRCCOPY );

		SelectObject( hMemDCsrc, hOldBitmap1 );
		SelectObject( hMemDCdst, hOldBitmap2 );
	}

	if( hOldPal1 )
		SelectPalette( hMemDCsrc, hOldPal1, FALSE );

	if( hOldPal2 )
		SelectPalette( hMemDCdst, hOldPal1, FALSE );

	DeleteDC( hMemDCsrc );
	DeleteDC( hMemDCdst );

	return( hNewBm );
}	// End - HBITMAP CropBitmap(HBITMAP, HPALETTE, LPRECT, LPPOINT )



//---------------------------------------------------------------------
//
// Function:   RenderFormat
//
// Purpose:    Renders an object for the clipboard.  The format is
//             specified in the "cf" variable (either CF_BITMAP,
//             CF_DIB, or CF_PALETTE).
//
//             This happens after a HandleCopyClipboard ...
//
//             Stolen almost verbatim out of ShowDIB.
//
// Parms:      hWndClip == Window clipboard belongs to, and where our
//                         image is stored).
//             cf       == Format to render (CF_BITMAP, CF_DIB, CF_PALETTE)
//                         Is actually the wParam of the WM_ message.
//             ptDIBSize== Size of the DIB in the given window.
//
// History:   Date      Reason
//             ???      Created
//             
//---------------------------------------------------------------------

HANDLE RenderFormat( HWND hWndClip, int cf, POINT ptDIBSize )
{
	HBITMAP   hBitmap;
//	HANDLE    hDIB;
	HANDLE    hDIBInfo;
	LPDIBINFO lpDIBInfo;
	HPALETTE  hPalette;        // Handle to the bitmap's palette.
	HANDLE    h = NULL;		// Start with NOTHING

	if( !hWndClip )		// Oops, NO ACTIVE WINDOW!!!
		return( h );

	hDIBInfo = GetWindowExtra( hWndClip, WW_DIB_HINFO );
	if( !hDIBInfo )		// Double Oops - NO INFORMATION!!!
		return( h );

	lpDIBInfo    = (LPDIBINFO) DVGlobalLock (hDIBInfo);
   if( !lpDIBInfo )     // triple oooops - NO LOCK HANDLE
      return( h );

//	hDIB         = lpDIBInfo->hDIB;
	hPalette     = lpDIBInfo->hPal;
	hBitmap      = lpDIBInfo->hBitmap;
	DVGlobalUnlock (hDIBInfo);
	switch( cf )
	{
	case CF_BITMAP:
		h = CropBitmap( hBitmap, hPalette, &grcClip, &ptDIBSize );
		break;

	case CF_DIB:
		{
			HBITMAP hbm;
         // NOTE:  For simplicity, we use the display device to crop
			//        the bitmap.  This means that we may lose color
			//        precision (if the display device has less color
			//        precision than the DIB).  This isn't usually a
			//        problem, as users shouldn't really be editting
			//        images on devices that can't display them.
			hbm = RenderFormat( hWndClip, CF_BITMAP, ptDIBSize );
			if( hbm )
			{
				h = DVBitmapToDIB( hbm, hPalette );
				DeleteObject( hbm );
			}
			break;
		}

	case CF_PALETTE:
		if( hPalette )
			h = CopyPaletteChangingFlags( hPalette, 0 );
		break;

	}

	return h;

}	// End - HANDLE RenderFormat( HWND hWndClip, int cf, POINT ptDIBSize )


//---------------------------------------------------------------------
//
// Function:   HandleCopyClipboard
//
// Purpose:    User wants to copy the current DIB to the clipboard.
//             Tell the clipboard we can render a DIB, DDB, and a
//             palette (defer rendering until we get a WM_RENDERFORMAT
//             in our MDI child window procedure in CHILD.C).
//             Handle IDM_COPY from MENU
//
// Parms:      None
//             
//
// History:   Date      Reason
//            6/1/91    Created
//           11/4/91    Init grcClip to full DIB size if
//                      it is currently empty.
//             
//---------------------------------------------------------------------

HWND  HandleCopyClipboard( void )
{
	HWND    hDIBWnd = GetCurrentMDIWnd();
	if( hDIBWnd )
   {
		// Clean clipboard of contents, and tell it we can render
		//  a DIB, a DDB, and/or a palette.
		if( OpenClipboard( hDIBWnd ) )
		{
			EmptyClipboard();
			SetClipboardData( CF_DIB,     NULL );
			SetClipboardData( CF_BITMAP,  NULL );
			SetClipboardData( CF_PALETTE, NULL );
			CloseClipboard ();

			// Set our globals to tell our app which child window
			//  owns the clipboard, and the clipping rectangle at
			//  the time of the copy.  If the clipping rectangle is
			//  empty, then use the entire DIB window.
			hWndClip   = hDIBWnd;
			grcClip    = GetCurrentClipRect( hWndClip );
			ptClipSize = GetCurrentDIBSize( hWndClip );
			if( IsRectEmpty( &grcClip ) )
			{
				grcClip.left   = 0;
				grcClip.top    = 0;
				grcClip.right  = ptClipSize.x;
				grcClip.bottom = ptClipSize.y;
			}
		}
		else
		{
			DIBError( ERR_CLIPBUSY );
         hDIBWnd = 0;
		}
   }
   else
	{
		DIBError( ERR_NOCLIPWINDOW );
	}

   return hDIBWnd;

}


//---------------------------------------------------------------------
//
// Function:   HandlePasteClipboard
//
// Purpose:    User wants to paste the clipboard's contents to our
//             app.  Open a new DIB window with the bitmap in the
//             clipboard.
//
// Parms:      None
//             
// History:   Date      Reason
//            6/1/91    Created
//             
//---------------------------------------------------------------------
HWND  HandlePasteClipboard( void )
{
   HWND        hWnd = 0;
	HANDLE      hDIB;
	HBITMAP     hBitmap;
	HPALETTE    hPal;
	PRDIB		   prd = (PRDIB)MALLOC( sizeof(RDIB) );
   if(!prd)
      chkme( "C:ERROR: No memory!!!"MEOR );
	NULPRDIB( prd );

   hDIB = 0;   // no DIB (device independant bitmap) yet

	// Open up the clipboard.  This routine assumes our app has
	//  the focus (which it should, as the user just picked the
	//  paste operation off the menu, and Windows is a non-preemptive
	//  system.  First we try for a DIB; if that's not available go
	//  for a bitmap (and a palette if one can be had).  Whatever
	//  format's available, we have to copy immediately (since the
	//  handle returned by GetClipboardData() belongs to the clipboard.
	//  Finally, go create the new MDI child window.
	if( OpenClipboard( GetFocus() ) )
	{
		if( IsClipboardFormatAvailable( CF_DIB ) )
		{
			hDIB = CopyHandle( GetClipboardData( CF_DIB ) );
		}
		else if( IsClipboardFormatAvailable( CF_BITMAP ) )
		{
			// HANDLE GetClipboardData(
			//    UINT uFormat );	// clipboard format  
			hBitmap = GetClipboardData( CF_BITMAP );

			if( IsClipboardFormatAvailable( CF_PALETTE ) )
				hPal = GetClipboardData( CF_PALETTE );
			else
				hPal = GetStockObject( DEFAULT_PALETTE );

			hDIB = DVBitmapToDIB( hBitmap, hPal );
		}
		else
		{
			DIBError( ERR_NOCLIPFORMATS );
		}

		CloseClipboard();

      if( !hDIB )
         goto Exit_Paste;

		prd->rd_pTitle = gszRPTit;     // set TITLE file name
      prd->rd_pPath  = gszRPNam;     // but keep the FULL NAME
		// The window title is of the form: "Clipboard1".  The
		//  number in the title is changed for each paste operation.
      // Get Next Name( &rd, &gszRPFrm[0], gnPasteNum++ );
      DVNextRDName2( prd, &gszRPFrm[0], &gnPasteNum, &gbChgPaste );   // = "TEMPB%03d.BMP"

		// Perform the actual window opening.
		prd->rd_hDIB    = hDIB;
		prd->rd_Caller  = df_CLIPBRD;

		//OpenDIBWindow( hDIB, szTitle, df_CLIPBRD );
		hWnd = OpenDIBWindow2( prd ); 
		if(hWnd)
      {
         if( WriteBMPFile2( gszRPNam, hDIB, TRUE ) )
         {
#ifdef	CHGADDTO
            //AddToFileList( szTitle );
            ADD2LIST( prd );
#endif	// CHGADDTO
         }
      }
	}
	else
	{
		DIBError (ERR_CLIPBUSY);
	}

Exit_Paste:

   MFREE(prd);

   return hWnd;

}	// End - void HandlePasteClipboard (void)

// eof
